@@ -3,6 +3,9 @@ use Carp ();
 
 ############################################################################
 package Net::OpenID::Association;
+BEGIN {
+  $Net::OpenID::Association::VERSION = '1.030099_001';
+}
 use fields (
             'server',    # author-identity identity server endpoint
             'secret',    # the secret for this association
@@ -13,6 +16,7 @@ use fields (
 
 use Storable ();
 use Digest::SHA1 qw(sha1);
+use Net::OpenID::Common;
 
 sub new {
     my Net::OpenID::Association $self = shift;
@@ -37,6 +41,12 @@ sub secret {
     $self->{'secret'};
 }
 
+sub type {
+    my $self = shift;
+    die if @_;
+    $self->{'type'};
+}
+
 sub server {
     my Net::OpenID::Association $self = shift;
     Carp::croak("Too many parameters") if @_;
@@ -63,7 +73,10 @@ sub usable {
 # goes into dumb consumer mode.  will do a POST and allocate
 # a new assoc_handle if none is found, or has expired
 sub server_assoc {
-    my ($csr, $server) = @_;
+    my ($csr, $server, $force_reassociate, %opts) = @_;
+
+    my $protocol_version = delete $opts{protocol_version} || 1;
+    Carp::croak("unknown options: " . join(", ", keys %opts)) if %opts;
 
     # closure to return undef (dumb consumer mode) and log why
     my $dumb = sub {
@@ -74,26 +87,32 @@ sub server_assoc {
     my $cache = $csr->cache;
     return $dumb->("no_cache") unless $cache;
 
-    # try first from cached association handle
-    if (my $handle = $cache->get("shandle:$server")) {
-        my $assoc = handle_assoc($csr, $server, $handle);
+    unless ($force_reassociate) {
+        # try first from cached association handle
+        if (my $handle = $cache->get("shandle:$server")) {
+            my $assoc = handle_assoc($csr, $server, $handle);
 
-        if ($assoc && $assoc->usable) {
-            $csr->_debug("Found association from cache (handle=$handle)");
-            return $assoc;
+            if ($assoc && $assoc->usable) {
+                $csr->_debug("Found association from cache (handle=$handle)");
+                return $assoc;
+            }
         }
     }
 
     # make a new association
-    my $dh = _default_dh();
+    my $dh = OpenID::util::get_dh();
 
     my %post = (
                 "openid.mode" => "associate",
                 "openid.assoc_type" => "HMAC-SHA1",
                 "openid.session_type" => "DH-SHA1",
-                "openid.dh_consumer_public" => OpenID::util::bi2arg($dh->pub_key),
+                "openid.dh_consumer_public" => OpenID::util::int2arg($dh->pub_key),
                 );
 
+    if ($protocol_version == 2) {
+        $post{"openid.ns"} = OpenID::util::version_2_namespace();
+    }
+
     my $req = HTTP::Request->new(POST => $server);
     $req->header("Content-Type" => "application/x-www-form-urlencoded");
     $req->content(join("&", map { "$_=" . OpenID::util::eurl($post{$_}) } keys %post));
@@ -139,9 +158,9 @@ sub server_assoc {
     if ($stype ne "DH-SHA1") {
         $secret = OpenID::util::d64($args{'mac_key'});
     } else {
-        my $server_pub = OpenID::util::arg2bi($args{'dh_server_public'});
+        my $server_pub = OpenID::util::arg2int($args{'dh_server_public'});
         my $dh_sec = $dh->compute_secret($server_pub);
-        $secret = OpenID::util::d64($args{'enc_mac_key'}) ^ sha1(OpenID::util::bi2bytes($dh_sec));
+        $secret = OpenID::util::d64($args{'enc_mac_key'}) ^ sha1(OpenID::util::int2bytes($dh_sec));
     }
     return $dumb->("secret_not_20_bytes") unless length($secret) == 20;
 
@@ -199,16 +218,6 @@ sub invalidate_handle {
     $cache->set("hassoc:$server:$handle", "");
 }
 
-sub _default_dh {
-    my $dh = Crypt::DH->new;
-    $dh->p("155172898181473697471232257763715539915724801966915404479707795314057629378541917580651227423698188993727816152646631438561595825688188889951272158842675419950341258706556549803580104870537681476726513255747040765857479291291572334510643245094715007229621094194349783925984760375594985848253359305585439638443");
-    $dh->g("2");
-    $dh->generate_keys;
-    return $dh;
-}
-
-
-
 1;
 
 __END__
@@ -217,6 +226,10 @@ __END__
 
 Net::OpenID::Association - a relationship with an identity server
 
+=head1 VERSION
+
+version 1.030099_001
+
 =head1 DESCRIPTION
 
 Internal class.
@@ -233,5 +246,4 @@ L<Net::OpenID::VerifiedIdentity>
 
 L<Net::OpenID::Server>
 
-Website:  L<http://www.danga.com/openid/>
-
+Website:  L<http://www.danga.com/openid/>
\ No newline at end of file
